#include "../misc/common.h"
#include "../misc/trace.h"
#include "../misc/timer.h"
#include "../misc/options.h"

#include "flow_sizing.h"

#define TOL 1e-6

RegisterOption(FD_step, "1", "step size for direction finding")
RegisterOption(line_epsi, "0.02", "expected increase for inexact line search")
RegisterOption(line_alpha, "2", "step ratio for inexact line search")
RegisterOption(line_inexact, "true", "use inexact line search")
RegisterOption(line_autostep, "false", "automatically determine FD step")
RegisterOption(tr_expand, "0.75", "ratio to expand trust region")
RegisterOption(tr_shrink, "0.25", "ratio to shrink trust region")

using misc::options;

double flow_sizing::compute_max_w() const
{
	double ub = 0;
	bool has_max = false;

	for (size_t i = 0; i < vars_->num_edges(); ++i)
	{
		if (df_[i] >= 0)
			continue;

		double w = -flow_[i]/df_[i];

		if (!has_max || (w < ub))
			ub = w;

		has_max = true;
	}

	if (has_max)
		return ub;
	else
		return -1;
}

flow_sizing::flow_sizing()
	: step_(1),
	s_(options::get().as_double("FD_step")),
	epsi_(options::get().as_double("line_epsi")),
	alpha_(options::get().as_double("line_alpha")),
	inexact_(options::get().as_bool("line_inexact")),
	autostep_(options::get().as_bool("line_autostep"))
{
}

bool flow_sizing::line_search_inexact()
{
	TraceScope("GTSIZING", 2);

	double dr0 = 0;
	
//	while (scale_base_ < 50000)
//	{
		dr0 = update_direction(s_);
//		if (dr0 > 0)
//			break;
//		scale_base_ *= 1.5;
//	}

	if (scale_base_ >= 50000)
	{
		fprintf(stdout, "  terminated by flow scaling %e\n", scale_base_);
		return false;
	}

	double cost0 = vars_->get_cost();

	TracePrint("  inexact start, s = %.3e, g' = %+.1e, g = %.3e (%.3e)\n",
		s_, dr0, cost0, before_LRS_);
	
	if (dr0 < TOL)
	{
		fprintf(stdout, "  terminated by derivative %e\n", dr0);
		return false;
	}
	
	double w_max = compute_max_w();

	if (autostep_)
		step_ = 1;

	if ((w_max >= 0) && (step_ > w_max))
		step_ = w_max;

	size_t iters = compute_LRS(step_);
	double dr = compute_derivative();
	double cost = vars_->get_cost();

	TracePrint("  inexact step0, w = %.3e, g' = %+.1e, g = %.3e (%.3e)\n"
		"    expect %.3e, iters %u, LRS_iters %u\n",
		step_, dr, cost, before_LRS_,
		cost0+dr0*step_*epsi_,
		iters, LRS_iters_);

	size_t i = 1;
	if (cost > cost0+dr0*step_*epsi_) // not too big, increase step
	{
		for (step_ *= alpha_; (w_max < 0) || (step_ <= w_max); step_ *= alpha_, ++i)
		{
			iters = compute_LRS(step_);
			dr = compute_derivative();
			cost = vars_->get_cost();

			TracePrint(
				"  inexact mul %u, w = %.3e, g' = %+.1e, g = %.3e (%.3e)\n"
				"    expect %.3e, iters %u, LRS_iters %u\n",
				i, step_, dr, cost, before_LRS_,
				cost0+dr0*step_*epsi_,
				iters, LRS_iters_);

			if (cost <= cost0+dr0*step_*epsi_) // big enough
				break;
		}
		step_ /= alpha_;
	}
	else // not too small, decrease step
	{
		for (step_ /= alpha_; step_ > TOL*sqrt(TOL); step_ /= alpha_, ++i)
		{
			iters = compute_LRS(step_);
			dr = compute_derivative();
			cost = vars_->get_cost();

			TracePrint(
				"  inexact div %u, w = %.3e, g' = %+.1e, g = %.3e (%.3e)\n"
				"    expect %.3e, iters %u, LRS_iters %u\n",
				i, step_, dr, cost, before_LRS_,
				cost0+dr0*step_*epsi_,
				iters, LRS_iters_);

			if (cost > cost0+dr0*step_*epsi_) // small enough
				break;
		}
	}

	if (autostep_)
	{
		if (step_ > 3)
			s_ *= 2;
		if (step_ < 0.3)
			s_ *= 0.5;

		if (s_ < sqrt(TOL))
			s_ = sqrt(TOL);
	}

	update_flow(step_);

	return true;
}

bool flow_sizing::line_search()
{
	TraceScope("GTSIZING", 2);

	double wl = 0;
	double drl = update_direction(s_);
	double costl = vars_->get_cost();

	TracePrint("  exact start, s = %.3e, g' = %+.1e, g = %.3e (%.3e)\n",
		s_, drl, costl, before_LRS_);
	
	if (drl < TOL)
	{
		fprintf(stdout, "  terminated by derivative %e\n", drl);
		return false;
	}
	
	double wr = compute_max_w();
	double drr = 0;
	double costr = 0;

	if (wr < 0) // not bounded by feasibility of flow
	{
		size_t i = 0;
		for (wr = 1;; wr *= 2, ++i)
		{
			size_t iters = compute_LRS(wr);
			drr = compute_derivative();
			costr = vars_->get_cost();

			TracePrint(
				"  max_w mul %u, w = %.3e, g' = %+.1e, g = %.3e (%.3e)\n"
				"    iters %u, LRS_iters %u\n",
				i, wr, drr, costr, before_LRS_,
				iters, LRS_iters_);

			if (drr < TOL)
				break;

			wl = wr, drl = drr, costl = costr;
		}
	}
	else
	{
		size_t iters = compute_LRS(wr);
		drr = compute_derivative();
		costr = vars_->get_cost();

		TracePrint(
			"  max_w step0, w = %.3e, g' = %+.1e, g = %.3e (%.3e)\n"
			"    iters %u, LRS_iters %u\n",
			wr, drr, costr, before_LRS_,
			iters, LRS_iters_);
	}

	if (drr < TOL) // max_w is decreasing
	{
		for (int i = 1;; ++i)
		{
			double wm = (wl+wr)/2;
			//double wm = (wr*drl-wl*drr)/(drl-drr);
	
			size_t iters = compute_LRS(wm);
	
			double drm = compute_derivative();
			double costm = vars_->get_cost();

			TracePrint(
				"  exact half %d, w = %.3e, g' = %+.1e, g = %.3e (%.3e)\n"
				"    range %.1e/%.1e, iters %u, total_iters %u\n",
				i, wm, drm, costm, before_LRS_,
				wr-wl, fabs(costr-costl),
				iters, LRS_iters_);

			if (drm >= 0)
				wl = wm, drl = drm, costl = costm;
			else
				wr = wm, drr = drm, costr = costm;

			if ((fabs(drr) < TOL)
				|| (fabs(drl) < TOL)
				|| (wr-wl < TOL)
				|| (fabs(costr-costl) < TOL))
				break;
		}
	}

	if (costr > costl)
		update_flow(wr);
	else
		update_flow(wl);

	return true;
}

bool flow_sizing::find_next_flow(int)
{
	if (inexact_)
		return line_search_inexact();
	else
		return line_search();
}

flow_sizing_tr::flow_sizing_tr()
	: s_(options::get().as_double("FD_step")),
	expand_(options::get().as_double("tr_expand")),
	shrink_(options::get().as_double("tr_shrink"))
{
}

bool flow_sizing_tr::find_next_flow(int)
{
	return trust_region_search();
}

bool flow_sizing_tr::trust_region_search()
{
	return false;
}
