#ifndef FLOW_SOLVER_H
#define FLOW_SOLVER_H

// this a flow solver class that will call CS2 directly
// we will move to our own code if necessary
// make sure you don't refer to CS2 types directly in your code
// then you can use our code when we remove the dependency to CS2

// got some types from CS2
extern "C" {
struct node_st;
struct arc_st;
}

namespace flow
{

class CS_solver
{
	// no copy, no assign
	CS_solver(CS_solver &);
	CS_solver &operator= (const CS_solver &);

	friend struct arc_tail;
	friend struct arc_swap;

public:
	CS_solver();
	CS_solver(size_t num_V, const size_t edges[], size_t num_E);
	void assign(size_t num_V, const size_t edges[], size_t num_E);
	
	// avoid compiling error on incomplte type for class clients
	~CS_solver();

	size_t num_vertices() const;
	size_t num_edges() const;

	// # nodes and arcs in the residue graph
	size_t num_nodes() const;
	size_t num_arcs() const;

	enum {CAP_MAX = 0x7fffffff, CAP_MIN = 0x80000000};

	void set_vertex_excess(size_t which_vertex, long long excess);
	void set_edge_cost(size_t which_edge, long long cost);
	void set_edge_caps(size_t which_edge, long lb, long ub); // will set flow to lb
	void set_edge_flow(size_t which_edge, long flow);

	long long get_vertex_excess(size_t which_vertex) const;
	long long get_edge_cost(size_t which_edge) const;
	long get_edge_lb(size_t which_edge) const;
	long get_edge_ub(size_t which_edge) const;
	long get_edge_flow(size_t which_edge) const;

	bool solve_unknown();
	bool solve_feasible();

	long long get_cost() const;

	void collect_edge_flows(long f[]) const;
	//void collect_vertex_potential(long p[]);

	bool backward_arc(size_t i_arc) const;
	bool forward_arc(size_t i_arc) const;

protected:

	typedef node_st node_t;
	typedef arc_st arc_t;

	std::vector<node_t> nodes_;
	std::vector<arc_t> arcs_;

	struct vertex_info
	{
		long long excess;

		vertex_info(long long exc = 0);

	}; // struct vertex_info

	struct edge_info
	{
		long long cost;
		long lb, ub, flow;

		edge_info(long l = 0, long u = 0, long f = 0, long long c = 0);

	}; // struct edge_info

	std::vector<vertex_info> vertex_infos_;
	std::vector<edge_info> edge_infos_;

	long long cost_;

	node_t default_node() const;
	arc_t default_arc() const;
	void init_arc_pair(size_t i_edge, size_t from, size_t to);

	long long initialize();
	void collect_everything();
	bool verify_optimal_price();

}; // class CS_solver

} // namespace flow

#endif // FLOW_SOLVER_H
