#include "../misc/common.h"
#include "../misc/token_parser.h"
#include "../misc/string_table.h"
#include "../misc/trace.h"
#include "../misc/auto_file.h"
#include "../misc/options.h"

#include "sizing_graph.h"

RegisterTraceScope(GTSIZING, 1, "trace level for gate sizing")

RegisterOption(scale, "1e9", "scale the delays, negative for automatic scaling")
RegisterOption(force_size_lb, "-1", "size lb, negative for file input")
RegisterOption(force_size_ub, "-1", "size ub, negative for file input")

RegisterOption(skew_opt, "false", "allow skew optimization")
RegisterOption(skew_lb, "0", "skew/T lb")
RegisterOption(skew_ub, "0.5", "skew/T ub")

sizing_graph::sizing_graph()
	: scale_(1), clock_period_(0),
	skew_lb_(0), skew_ub_(0.5),
	skew_opt_(false)
{
}

const std::string &sizing_graph::get_graph_name() const
{
	return graph_name_;
}

double sizing_graph::get_scale() const
{
	return scale_;
}

size_t sizing_graph::num_vars() const
{
	return var_names_.size();
}

size_t sizing_graph::num_vertices() const
{
	return vertex_names_.size();
}

size_t sizing_graph::num_edges() const
{
	return edge_delays_.size();
}

const misc::cstr_t *sizing_graph::get_var_names() const
{
	return &var_names_[0];
}

const misc::cstr_t *sizing_graph::get_vertex_names() const
{
	return &vertex_names_[0];
}

const double *sizing_graph::get_var_weights() const
{
	return &var_weights_[0];
}

const size_t *sizing_graph::get_edges() const
{
	return &edges_[0];
}

double sizing_graph::get_var_lb(size_t which_var) const
{
	assert(which_var < num_vars());

	return var_bounds_[which_var*2];
}

double sizing_graph::get_var_ub(size_t which_var) const
{
	assert(which_var < num_vars());

	return var_bounds_[which_var*2+1];
}

double sizing_graph::get_var_weight(size_t which_var) const
{
	assert(which_var < num_vars());

	return var_weights_[which_var];
}

size_t sizing_graph::get_edge_from(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return edges_[which_edge*2];
}

size_t sizing_graph::get_edge_to(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return edges_[which_edge*2+1];
}

const math::posynomial &sizing_graph::get_constraint(size_t which_constraint) const
{
  assert(which_constraint < num_packing_constraints() );
  
  return grid_constraints_[which_constraint];
}

double sizing_graph::get_constraint_ub(size_t which_constraint) const
{
  assert(which_constraint < num_packing_constraints() );
  
  return grid_bounds_[which_constraint];
}

const math::posynomial &sizing_graph::get_edge_delay(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return edge_delays_[which_edge];
}

sizing_graph::id_crange sizing_graph::get_related_edges(size_t which_var) const
{
	assert(which_var < num_vars());

	return id_crange(
		var_edges_[which_var].begin(),
		var_edges_[which_var].end());
}

sizing_graph::id_crange sizing_graph::get_related_vars(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return id_crange(
		edge_vars_[which_edge].begin(),
		edge_vars_[which_edge].end());
}

sizing_graph::edge_range sizing_graph::fanins(size_t which_vertex) const
{
	return graph_.fanins(graph2::graph_v::vertex_descriptor(which_vertex));
}

sizing_graph::edge_range sizing_graph::fanouts(size_t which_vertex) const
{
	return graph_.fanouts(graph2::graph_v::vertex_descriptor(which_vertex));
}

size_t sizing_graph::this_edge(out_edge_iterator oi) const
{
	return graph_.this_edge(oi).id();
}

/*
# vertices
vertex_name arrival_time
# vars
weight lb ub var_name
# edges
from to delay_posynomial
*/
bool sizing_graph::load_cost(const std::string &graph_name)
{
  TraceScope("GTSIZING", 1);
  
  os::auto_file af;
  
  if (!af.file_open(graph_name, "r"))
    {
      fprintf(stderr, "%s: error: cannot read\n", graph_name.c_str());
      return false;
    }

  FILE *fp = af.get_fp();
  
  cost_.load(fp);

  size_t n = 0;
  
  if ( fscanf(fp, "%lu", &n) != 1 )
    {
      fprintf(stderr, "%s: error: bad # of vertices\n",
	      graph_name_.c_str());
      return false;
    }
  grid_bounds_.resize(n);
  grid_constraints_.resize(n);

  for(size_t i = 0; i < n; i++)
    {
      double ub;
      fscanf(fp, "%lf", &ub);
      grid_bounds_[i] = ub;
      grid_constraints_[i].load(fp);
    }
  return true;
}

bool sizing_graph::load_constraints(const std::string &const_file)
{
  TraceScope("GTSIZING", 1);
  
  os::auto_file af;
  
  if (!af.file_open(const_file, "r"))
    {
      fprintf(stderr, "%s: error: cannot read\n", const_file.c_str());
      return false;
    }
  
  FILE *fp = af.get_fp();
  
  size_t n = 0;
  
  if ( fscanf(fp, "%lu", &n) != 1 )
    {
      fprintf(stderr, "%s: error: bad # of vertices\n",
	      graph_name_.c_str());
      return false;
    }
  grid_bounds_.resize(n);
  grid_constraints_.resize(n);

  for(size_t i = 0; i < n; i++)
    {
      double ub;
      fscanf(fp, "%lf", &ub);
      grid_bounds_[i] = ub;
      grid_constraints_[i].load(fp);
    }
  return true;
}

double sizing_graph::compute_cost(const double * vars) const
{
  double cost = cost_.evaluate(vars);
  return cost;
}

bool sizing_graph::load(const std::string &graph_name)
{
	TraceScope("GTSIZING", 1);

	graph_name_ = graph_name;

	os::auto_file af;

	if (!af.file_open(graph_name, "r"))
	{
		fprintf(stderr, "%s: error: cannot read\n", graph_name.c_str());
		return false;
	}

	FILE *fp = af.get_fp();

	if (!load_vertices(fp))
		return false;

	if (!load_vars(fp))
		return false;

	if (!load_edges(fp))
		return false;

	graph_.assign(num_vertices(), &edges_[0], num_edges());

	update_scale();

	if (!detect_single_clock_period_edge())
		return false;

	mark_clock_period_edges();

	if (!detect_deb_dffs())
		return false;

	skew_opt_ = misc::options::get().as_bool("skew_opt");
	skew_lb_ = misc::options::get().as_double("skew_lb");
	skew_ub_ = misc::options::get().as_double("skew_ub");

	if (skew_opt_)
		expand_graph_for_skew_opt();

	TracePrint("%s: %u vertices, %u vars, %u edges, %u dffs\n"
		"  scale %.3e, clock period %.3f\n",
		get_graph_name().c_str(), num_vertices(), num_vars(),
		num_edges(), num_dffs(), get_scale(), get_clock_period());

	{
		TraceScope("GTSIZING", 6);

		if (TraceActive())
			dump(stdout, ".3f");
	}

	return true;
}

void sizing_graph::update_scale()
{
	scale_ = misc::options::get().as_double("scale");

	if (scale_ < 0)
	{
		double s[3];

		s[0] = compute_scale(0);
		s[1] = compute_scale(0.5);
		s[2] = compute_scale(1);

		scale_ = *std::max_element(s, s+3);
	}

	for (size_t i = 0; i < num_edges(); ++i)
	{
		typedef math::posynomial::sum_term_range sum_term_range;
		typedef math::posynomial::product_terms product_terms;

		for (sum_term_range r = edge_delays_[i].get_term_range();
			r.first != r.second; ++r.first)
			r.first->sum_coef *= scale_;
	}
}

double sizing_graph::compute_scale(double a)
{
	TraceScope("GTSIZING", 2);

	double cost = 0;
	double delay = 0;

	std::vector<double> vars(num_vars());

	for (size_t i = 0;i < num_vars(); ++i)
	{
		double new_var = get_var_lb(i)*a+get_var_ub(i)*(1-a);

		vars[i] = new_var;
		cost += new_var*get_var_weight(i);
	}

	for (size_t i = 0; i < num_edges(); ++i)
		delay += fabs(get_edge_delay(i).evaluate(&vars[0]));

	double scale = cost/delay;

	TracePrint("scaling: a %.3f, cost %.1e, delay %.1e, scale %.3e\n",
		a, cost, delay, scale);

	return scale;
}

bool sizing_graph::load_vertices(FILE *fp)
{
	size_t n = 0;

	if (fscanf(fp, "%lu", &n) != 1)
	{
		fprintf(stderr, "%s: error: bad # of vertices\n",
			graph_name_.c_str());
		return false;
	}

	vertex_names_.reserve(n);

	for (size_t i = 0; i < n; ++i)
	{
		char name[1000];
		double at;
		if (fscanf(fp, "%s %lf", name, &at) != 2)
		{
			fprintf(stderr, "%s: error: bad name for vertex %lu\n",
				graph_name_.c_str(), i);
			return false;
		}

		vertex_names_.push_back(misc::string_table_store_str(name));
	}

	return true;
}

bool sizing_graph::load_vars(FILE *fp)
{
	size_t n = 0;

	if (fscanf(fp, "%lu", &n) != 1)
	{
		fprintf(stderr, "%s: error: bad # of variables\n",
			graph_name_.c_str());
		return false;
	}

	var_names_.reserve(n);
	var_weights_.reserve(n);
	var_bounds_.reserve(2*n);
	
	var_edges_.resize(n);
	
	double force_lb = misc::options::get().as_double("force_size_lb");
	double force_ub = misc::options::get().as_double("force_size_ub");

	for (size_t i = 0; i < n; ++i)
	{
		double w = 0, lb = 0, ub = 0;
		if (fscanf(fp, "%lf %lf %lf", &w, &lb, &ub) != 3)
		{
			fprintf(stderr, "%s: error: bad params for variable %lu\n",
				graph_name_.c_str(), i);
			return false;
		}

		char name[1000];
		if (fscanf(fp, "%s", name) != 1)
		{
			fprintf(stderr, "%s: error: bad name for variable %lu\n",
				graph_name_.c_str(), i);
			return false;
		}

		var_names_.push_back(misc::string_table_store_str(name));
		var_weights_.push_back(w);

		var_bounds_.push_back((force_lb < 0)? lb: force_lb);
		var_bounds_.push_back((force_ub < 0)? ub: force_ub);
	}

	return true;
}

bool sizing_graph::load_edges(FILE *fp)
{
	size_t n = 0;

	if (fscanf(fp, "%lu", &n) != 1)
	{
		fprintf(stderr, "%s: error: bad # of edges\n",
			graph_name_.c_str());
		return false;
	}

	edges_.reserve(2*n);

	edge_delays_.resize(n);
	edge_vars_.resize(n);
	
	for (size_t i = 0; i < n; ++i)
	{
		size_t from, to;
		
		if (fscanf(fp, "%lu %lu", &from, &to) != 2)
		{
			fprintf(stderr, "%s: error: bad params for edge %lu\n",
				graph_name_.c_str(), i);
			return false;
		}
		
		if (!edge_delays_[i].load(fp))
		{
			fprintf(stderr, "%s: error: bad delay for edge %lu\n",
				graph_name_.c_str(), i);
			return false;
		}
		
		/*if ( misc::options::get().as_bool("temperature") == true ) {
		  double teffect;
		  fscanf(fp, "%lf", &teffect);
		  if ( teffect > 0 ) {
		    math::posynomial p;
		    p.add_term(teffect);
		    edge_delays_[i] += p;
		  }
		  }*/
		
		assert(from < num_vertices());
		assert(to < num_vertices());

		edges_.push_back(from);
		edges_.push_back(to);
		
		typedef math::posynomial::sum_term_range sum_term_range;
		typedef math::posynomial::product_terms product_terms;
		for (sum_term_range r = edge_delays_[i].get_term_range();
			r.first != r.second; ++r.first)
		{
			product_terms &terms = r.first->prod_terms;

			for (size_t j = 0; j < terms.size(); ++j)
			{
				assert(terms[j].var_name < num_vars());

				edge_vars_[i].insert(terms[j].var_name);
				var_edges_[terms[j].var_name].insert(i);
			}
		}
	}
	
	return true;
}

misc::cstr_t sizing_graph::get_vertex_name(size_t which_vertex) const
{
	assert(which_vertex < num_vertices());

	return vertex_names_[which_vertex];
}

misc::cstr_t sizing_graph::get_var_name(size_t which_var) const
{
	assert(which_var < num_vars());

	return var_names_[which_var];
}

void sizing_graph::dump(FILE *fp, misc::cstr_t format)
{
	fprintf(fp,
		"+++++++++++++++++++++++++++++++++++++++++++++\n"
		"%s: %lu vertices, %lu vars, %lu edges\n",
		graph_name_.c_str(),
		num_vertices(),
		num_vars(),
		num_edges());

	for (size_t i = 0; i < num_vertices(); ++i)
		fprintf(fp, "vertex %lu: %s\n", i, get_vertex_name(i));

	for (size_t i = 0; i < num_edges(); ++i)
		fprintf(fp,
			"edge %lu: %lu->%lu, %s\n",
			i,
			get_edge_from(i),
			get_edge_to(i),
			get_edge_delay(i).to_string(get_var_names(), format).c_str());

	fprintf(fp,
		"---------------------------------------------\n");
}

size_t sizing_graph::num_clock_period_edges() const
{
	return cpes_.size();
}

size_t sizing_graph::get_clock_period_edge(size_t which_cpe) const
{
	assert(which_cpe < num_clock_period_edges());

	return cpes_[which_cpe];
}

double sizing_graph::get_clock_period() const
{
	return clock_period_;
}

/*
void sizing_graph::set_clock_period(double cp)
{
	assert(cp > 0);

	for (size_t i = 0; i < num_clock_period_edges(); ++i)
	{
		size_t cpe = get_clock_period_edge(i);

		math::posynomial &d = edge_delays_[cpe];

		assert(d.num_terms() == 1);

		math::posynomial::sum_term &st = d.get_terms().front();

		assert(st.prod_terms.empty());

		assert(fabs((st.sum_coef+clock_period_)/clock_period_) < 1e-6);

		st.sum_coef = -cp;
	}

	clock_period_ = cp;
}
//*/

bool sizing_graph::detect_single_clock_period_edge()
{
	assert(cpes_.empty());

	for (size_t i = 0; i < num_edges(); ++i)
	{
		math::posynomial &cp = edge_delays_[i];

		if (cp.num_terms() != 1)
			continue;

		math::posynomial::sum_term &st = cp.get_terms().front();

		if (!st.prod_terms.empty())
			continue;

		if (st.sum_coef >= 0)
			continue;

		if (cpes_.empty())
		{
			cpes_.push_back(i);
			clock_period_ = -st.sum_coef;
			fprintf(stdout,
				"found clock period edge %lu (%lu->%lu) with clock period %e\n",
				i, get_edge_from(i), get_edge_to(i), clock_period_);
		}
		else
		{
			fprintf(stderr,
				"%s: error: >1 negative delay %lu (%lu->%lu) with %e\n",
				graph_name_.c_str(),
				i, get_edge_from(i), get_edge_to(i), -st.sum_coef);

			return false;
		}
	}

	return true;
}

bool sizing_graph::is_clock_period_edge(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return cpflags_[which_edge];
}

void sizing_graph::mark_clock_period_edges()
{
	cpflags_.assign(num_edges(), false);

	for (size_t i = 0; i < num_clock_period_edges(); ++i)
		cpflags_[get_clock_period_edge(i)] = true;
}

bool sizing_graph::is_skew_opt() const
{
	return skew_opt_;
}

size_t sizing_graph::num_dffs() const
{
	assert(dffs_.size()%2 == 0);

	return dffs_.size()/2;
}

size_t sizing_graph::get_Q_vertex(size_t which_dff) const
{
	assert(which_dff < num_dffs());

	return dffs_[which_dff];
}

size_t sizing_graph::get_D_vertex(size_t which_dff) const
{
	assert(which_dff < num_dffs());

	return dffs_[which_dff+num_dffs()];
}

size_t sizing_graph::get_IQ_edge(size_t which_dff) const
{
	assert(which_dff < num_dffs());

	return dff_edges_[which_dff];
}

size_t sizing_graph::get_DO_edge(size_t which_dff) const
{
	assert(which_dff < num_dffs());

	return dff_edges_[which_dff+num_dffs()];
}

size_t sizing_graph::get_DQ_edge(size_t which_dff) const
{
	assert(which_dff < num_dffs());

	return num_edges()-num_dffs()+which_dff;
}

bool sizing_graph::detect_deb_dffs()
{
	size_t I = 0;
	size_t O = num_vertices()-1;

	if (num_clock_period_edges() != 1)
	{
		fprintf(stderr,
			"%s: error: deb assumes exactly one clock period edge\n",
			graph_name_.c_str());

		return false;
	}

	size_t OI = get_clock_period_edge(0);

	if ((get_edge_from(OI) != O) && (get_edge_to(OI) != I))
	{
		fprintf(stderr,
			"%s: error: deb assumes I/O are first/last vertices\n",
			graph_name_.c_str());

		return false;
	}

	dffs_.reserve(10000);
	dff_edges_.reserve(10000);

	std::vector<std::string> tokens;

	std::map<std::string, size_t> dff_ids;

	// find Qs
	for (edge_range r = fanouts(I); r.first != r.second; ++r.first)
	{
		size_t IQ = this_edge(r.first);
		size_t Q = get_edge_to(IQ);

		std::string Q_name = get_vertex_name(Q);

		tokens.resize(0);
		misc::split_tokens(get_vertex_name(Q), &tokens, "/");

		if (tokens.size() == 1) // PI
			continue;

		if (tokens.size() != 2)
		{
			fprintf(stderr,
				"%s: error: deb assumes gate/port for vertex name, but found %s\n",
				graph_name_.c_str(), get_vertex_name(Q));

			return false;
		}

		if (tokens[1] != "0")
		{
			fprintf(stdout,
				"%s: warning: possible D port at vertex %s\n",
				graph_name_.c_str(), get_vertex_name(Q));

			continue;
		}

		if (dff_ids.find(tokens[0]) != dff_ids.end())
		{
			fprintf(stderr,
				"%s: error: multiple Q ports at vertex %s\n",
				graph_name_.c_str(), get_vertex_name(Q));

			return false;
		}

		dff_ids[tokens[0]] = dffs_.size();
		dffs_.push_back(Q);
		dff_edges_.push_back(IQ);
	}

	dffs_.resize(dffs_.size()*2);
	dff_edges_.resize(dff_edges_.size()*2);
	size_t D_count = 0;

	// find Ds
	for (edge_range r = fanins(O); r.first != r.second; ++r.first)
	{
		size_t DO = this_edge(r.first);
		size_t D = get_edge_from(DO);

		std::string D_name = get_vertex_name(D);

		tokens.resize(0);
		misc::split_tokens(get_vertex_name(D), &tokens, "/");

		if (tokens.size() == 1) // PI
			continue;

		if (tokens.size() != 2)
		{
			fprintf(stderr,
				"%s: error: deb assumes gate/port for vertex name, but found %s\n",
				graph_name_.c_str(), get_vertex_name(D));

			return false;
		}

		if (tokens[1] != "1") // PO
			continue;

		if (dff_ids.find(tokens[0]) == dff_ids.end())
		{
			fprintf(stderr,
				"%s: error: missing Q port at vertex %s\n",
				graph_name_.c_str(), get_vertex_name(D));

			return false;
		}

		++D_count;
		size_t dff = dff_ids[tokens[0]];
		dff_ids.erase(tokens[0]);
		dffs_[dff+num_dffs()] = D;
		dff_edges_[dff+num_dffs()] = DO;
	}

	if (D_count < num_dffs())
	{
		fprintf(stderr,
			"%s: error: missing %lu D ports\n",
			graph_name_.c_str(), num_dffs()-D_count);

		return false;
	}

	return true;
}

void sizing_graph::expand_graph_for_skew_opt()
{
	assert(skew_opt_);

	fprintf(stdout, "modify graph for skew optimization, add %lu edges\n", num_dffs());

	math::posynomial dIQ, dDO, dDQ;

	dIQ.add_term(get_clock_period()*skew_lb_);
	dDO.add_term(-get_clock_period()*skew_ub_);
	dDQ.add_term(-get_clock_period());

	edges_.reserve(edges_.size()+num_dffs()*2);
	edge_delays_.reserve(edge_delays_.size()+num_dffs());
	edge_vars_.reserve(edge_vars_.size()+num_dffs());
	cpes_.reserve(cpes_.size()+num_dffs());
	cpflags_.reserve(cpflags_.size()+num_dffs());

	for (size_t i = 0; i < num_dffs(); ++i)
	{
		edge_delays_[get_IQ_edge(i)] = dIQ;
		edge_delays_[get_DO_edge(i)] = dDO;

		cpes_.push_back(num_edges());
		cpflags_.push_back(true);
		edges_.push_back(get_D_vertex(i));
		edges_.push_back(get_Q_vertex(i));
		edge_delays_.push_back(dDQ);
		edge_vars_.push_back(id_set());
	}

	graph_.assign(num_vertices(), &edges_[0], num_edges());
}
