#ifndef SIZING_GRAPH_H
#define SIZING_GRAPH_H

#include "posynomial.h"
#include "../graph/graph2_v.h"

class sizing_graph
{
public:
	sizing_graph();

	size_t num_vars() const;
	size_t num_vertices() const;
	size_t num_edges() const;

	const std::string &get_graph_name() const;

	double get_scale() const;

	const misc::cstr_t *get_var_names() const;
	const misc::cstr_t *get_vertex_names() const;
	const double *get_var_weights() const;
	const size_t *get_edges() const;

	misc::cstr_t get_vertex_name(size_t which_vertex) const;
	misc::cstr_t get_var_name(size_t which_var) const;
	
	double get_var_lb(size_t which_var) const;
	double get_var_ub(size_t which_var) const;
	double get_var_weight(size_t which_var) const;

	size_t get_edge_from(size_t which_edge) const;
	size_t get_edge_to(size_t which_edge) const;
	const math::posynomial &get_edge_delay(size_t which_edge) const;

	typedef std::set<size_t> id_set;
	typedef id_set::const_iterator id_citer;
	typedef std::pair<id_citer, id_citer> id_crange;

	id_crange get_related_edges(size_t which_var) const;
	id_crange get_related_vars(size_t which_edge) const;

	typedef graph2::graph_v::out_edge_iterator out_edge_iterator;
	typedef graph2::graph_v::edge_range edge_range;

	edge_range fanins(size_t which_vertex) const;
	edge_range fanouts(size_t which_vertex) const;
	size_t this_edge(out_edge_iterator oi) const;

	bool load(const std::string &graph_name);
	
	//Function to load cost function and constraints
	bool load_cost(const std::string &sg_name);
	bool load_constraints(const std::string &const_file);
	size_t num_cost_terms() const {
	  return cost_.num_terms();
	}
	size_t num_packing_constraints() const {
	  return grid_constraints_.size();
	}
	double compute_cost(const double * vars) const;
	const math::posynomial &get_constraint(size_t which_constraint) const;
	double get_constraint_ub(size_t which_constraint) const;
	const math::posynomial get_cost_function() const {
	  return cost_;
	}
	
	void dump(FILE *fp, misc::cstr_t format = "e");

	size_t num_clock_period_edges() const;
	size_t get_clock_period_edge(size_t which_cpe) const;
	bool is_clock_period_edge(size_t which_edge) const;

	// note currently we assume
	// 1 skew upper-bounds are absolute;
	// 2 rat are relative to the clock period,
	// thus if clock period changes we only update
	// delays for clock period edges
	// void set_clock_period(double cp);

	double get_clock_period() const;

	bool is_skew_opt() const;

	size_t num_dffs() const;
	size_t get_Q_vertex(size_t which_dff) const;
	size_t get_D_vertex(size_t which_dff) const;
	size_t get_IQ_edge(size_t which_dff) const;
	size_t get_DO_edge(size_t which_dff) const;
	size_t get_DQ_edge(size_t which_dff) const;
	double get_skew_lb() const {return skew_lb_;}
	double get_skew_ub() const {return skew_ub_;}

	const graph2::graph_v *get_graph() const {return &graph_;}

protected:

	void update_scale();

	// compute total_cost/total_delay for a*lb+(1-a)*ub
	double compute_scale(double a);
	
	std::vector<misc::cstr_t> vertex_names_;
	//variables on each vertex to keep track of arrival times
	//std::vector<double> vertex_vars_;

	std::vector<misc::cstr_t> var_names_;
	std::vector<double> var_weights_;
	std::vector<double> var_bounds_; // lb, ub
	std::vector<id_set> var_edges_;
	
	//Debasish Das begin
	//cost and constraints needed for temperature optimization
	std::vector<math::posynomial> grid_constraints_;
	std::vector<double> grid_bounds_;
	math::posynomial cost_;
	//end
	
	std::vector<size_t> edges_; // from, to
	std::vector<math::posynomial> edge_delays_;
	std::vector<id_set> edge_vars_;
	
	graph2::graph_v graph_; // stores the adjacency structure

	std::string graph_name_;
	double scale_;

	bool load_vertices(FILE *fp);
	bool load_vars(FILE *fp);
	bool load_edges(FILE *fp);
	
	// the special clock period edges
	std::vector<size_t> cpes_;
	double clock_period_;
	std::vector<bool> cpflags_;

	bool detect_single_clock_period_edge();
	void mark_clock_period_edges();

	double skew_lb_, skew_ub_;
	bool skew_opt_;
	std::vector<size_t> dffs_, dff_edges_; // Q, D

	bool detect_deb_dffs();

	void expand_graph_for_skew_opt();

}; // class sizing_graph

#endif // SIZING_GRAPH_H
