#include "sparse.h"

NRsparseMat::NRsparseMat() : nrows(0), ncols(0), nvals(0), col_ptr(),
			     row_ind(), val() {}

NRsparseMat::NRsparseMat(Int m,Int n,Int nnvals) : 
  nrows(m),ncols(n),nvals(nnvals),col_ptr(n+1,0),
  row_ind(nnvals,0),val(nnvals,0.0) {}

VecDoub NRsparseMat::ax(const VecDoub &x) const {
  VecDoub y(nrows,0.0);
  for (Int j=0;j<ncols;j++) {
    for (Int i=col_ptr[j];i<col_ptr[j+1];i++)
      y[row_ind[i]] += val[i]*x[j];
  }
  return y;
}

VecDoub NRsparseMat::atx(const VecDoub &x) const {
  VecDoub y(ncols);
  for (Int i=0;i<ncols;i++) {
    y[i]=0.0;
    for (Int j=col_ptr[i];j<col_ptr[i+1];j++)
      y[i] += val[j]*x[row_ind[j]];
  }
  return y;
}

NRsparseMat NRsparseMat::transpose() const {
  Int i,j,k,index,m=nrows,n=ncols;
  NRsparseMat at(n,m,nvals);
  VecInt count(m,0);
  for (i=0;i<n;i++)
    for (j=col_ptr[i];j<col_ptr[i+1];j++) {
      k=row_ind[j];
      count[k]++;
    }
  for (j=0;j<m;j++)
    at.col_ptr[j+1]=at.col_ptr[j]+count[j];
  for(j=0;j<m;j++)
    count[j]=0;
  for (i=0;i<n;i++)
    for (j=col_ptr[i];j<col_ptr[i+1];j++) {
      k=row_ind[j];
      index=at.col_ptr[k]+count[k];
      at.row_ind[index]=i;
      at.val[index]=val[j];
      count[k]++;
    }
  return at;
}

bool NRsparseMat::generateDense(MatDoub& d)
{
  if ( d.nrows() != nrows && d.ncols() != ncols )
    return false;
  
  for(int i = 0; i < nrows; i++)
    for(int j = 0; j < ncols; j++)
      d[i][j] = 0.0;
  
  for(int i = 0; i < ncols; i++)
    for(int j = col_ptr[i]; j < col_ptr[i+1]; j++){
      int k = row_ind[j];
      d[k][i] = val[j];
    }
  return true;
}

void NRsparseMat::save(FILE *fp) const {
  fprintf(fp, "%d %d %d\n", nrows, ncols, nvals);
  for(int i = 0; i < val.size(); i++)
    fprintf(fp, "%g ", val[i]);
  fprintf(fp,"\n");
  for(int i = 0; i < row_ind.size(); i++)
    fprintf(fp, "%d ", row_ind[i]);
  fprintf(fp,"\n");
  for(int i = 0; i < col_ptr.size(); i++)
    fprintf(fp, "%d ", col_ptr[i]);
  fprintf(fp,"\n");
}

bool NRsparseMat::load(FILE *fp) {
  if ( fscanf(fp, "%d %d %d", &nrows, &ncols, &nvals) != 3 )
    return false;
  printf("rows : %d cols : %d nvals : %d\n", nrows, ncols, nvals);
  
  col_ptr.resize(ncols+1);
  row_ind.resize(nvals);
  val.resize(nvals);
  
  double v;
  int index;
  
  for(int i = 0; i < nvals; i++) {
    fscanf(fp, "%lf", &v);
    val[i] = v;
  }
  
  for(int i = 0; i < nvals; i++){
    fscanf(fp, "%d", &index);
    row_ind[i] = index;
  }
  
  for(int i = 0; i < ncols+1; i++){
    fscanf(fp, "%d", &index);
    col_ptr[i] = index;
  }
  
  return true;
}

void Linbcg::solve(VecDoub_I &b, VecDoub_IO &x, const Int itol, const Doub tol,
	const Int itmax, Int &iter, Doub &err)
{
	Doub ak,akden,bk,bkden=1.0,bknum,bnrm,dxnrm,xnrm,zm1nrm,znrm;
	const Doub EPS=1.0e-14;
	Int j,n=b.size();
	VecDoub p(n),pp(n),r(n),rr(n),z(n),zz(n);
	iter=0;
	atimes(x,r,0);
	for (j=0;j<n;j++) {
		r[j]=b[j]-r[j];
		rr[j]=r[j];
	}
	//atimes(r,rr,0);
	if (itol == 1) {
		bnrm=snrm(b,itol);
		asolve(r,z,0);
	}
	else if (itol == 2) {
		asolve(b,z,0);
		bnrm=snrm(z,itol);
		asolve(r,z,0);
	}
	else if (itol == 3 || itol == 4) {
		asolve(b,z,0);
		bnrm=snrm(z,itol);
		asolve(r,z,0);
		znrm=snrm(z,itol);
	} else throw("illegal itol in linbcg");
	while (iter < itmax) {
		++iter;
		asolve(rr,zz,1);
		for (bknum=0.0,j=0;j<n;j++) bknum += z[j]*rr[j];
		if (iter == 1) {
			for (j=0;j<n;j++) {
				p[j]=z[j];
				pp[j]=zz[j];
			}
		} else {
			bk=bknum/bkden;
			for (j=0;j<n;j++) {
				p[j]=bk*p[j]+z[j];
				pp[j]=bk*pp[j]+zz[j];
			}
		}
		bkden=bknum;
		atimes(p,z,0);
		for (akden=0.0,j=0;j<n;j++) akden += z[j]*pp[j];
		ak=bknum/akden;
		atimes(pp,zz,1);
		for (j=0;j<n;j++) {
			x[j] += ak*p[j];
			r[j] -= ak*z[j];
			rr[j] -= ak*zz[j];
		}
		asolve(r,z,0);
		if (itol == 1)
			err=snrm(r,itol)/bnrm;
		else if (itol == 2)
			err=snrm(z,itol)/bnrm;
		else if (itol == 3 || itol == 4) {
			zm1nrm=znrm;
			znrm=snrm(z,itol);
			if (abs(zm1nrm-znrm) > EPS*znrm) {
				dxnrm=abs(ak)*snrm(p,itol);
				err=znrm/abs(zm1nrm-znrm)*dxnrm;
			} else {
				err=znrm/bnrm;
				continue;
			}
			xnrm=snrm(x,itol);
			if (err <= 0.5*xnrm) err /= xnrm;
			else {
				err=znrm/bnrm;
				continue;
			}
		}
		if (err <= tol) break;
	}
}

Doub Linbcg::snrm(VecDoub_I &sx, const Int itol)
{
	Int i,isamax,n=sx.size();
	Doub ans;
	if (itol <= 3) {
		ans = 0.0;
		for (i=0;i<n;i++) ans += SQR(sx[i]);
		return sqrt(ans);
	} else {
		isamax=0;
		for (i=0;i<n;i++) {
			if (abs(sx[i]) > abs(sx[isamax])) isamax=i;
		}
		return abs(sx[isamax]);
	}
}

void NRsparseLinbcg::atimes(VecDoub_I &x, VecDoub_O &r, const Int itrnsp) 
{
  if (itrnsp) r=mat.atx(x);
  else r=mat.ax(x);
}

void NRsparseLinbcg::asolve(VecDoub_I &b, VecDoub_O &x, const Int itrnsp) 
{
  Int i,j;
  Doub diag;
  for (i=0;i<n;i++) {
    diag=0.0;
    for (j=mat.col_ptr[i];j<mat.col_ptr[i+1];j++)
      if (mat.row_ind[j] == i) {
	diag=mat.val[j];
	break;
      }
    x[i]=(diag != 0.0 ? b[i]/diag : b[i]);
  }
}
