/*
 * Copyright (c) 2003-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * graph2_edge.h - Interface for edge management.
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#ifndef GRAPH2_EDGE_H
#define GRAPH2_EDGE_H

#include "graph2_property.h"

namespace graph2
{

namespace detail
{

class edge_descriptor_impl
{
	size_t e_;
public:
	explicit edge_descriptor_impl(size_t e) : e_(e) {}
	edge_descriptor_impl() : e_(size_t(-1)) {}
	size_t id() const {return e_;}

	bool operator == (edge_descriptor_impl e) const {return e.e_ == e_;}

	bool operator < (edge_descriptor_impl e) const {return e_ < e.e_;}
	bool operator > (edge_descriptor_impl e) const {return e_ > e.e_;}

	// just like those in vertex_descriptor
	edge_descriptor_impl& operator ++ () {++e_; return *this;}
	edge_descriptor_impl& operator -- () {--e_; return *this;}
}; // class edge_descriptor_impl

class edge_set_base
{
protected:
	size_t num_edges_;

	edge_set_base(size_t n = 0) : num_edges_(n) {}

public:
	typedef edge_descriptor_impl edge_descriptor;

	size_t num_edges() const {return num_edges_;}
	bool valid_edge(edge_descriptor e) const {return e.id() < num_edges();}

	// we need a concrete type for edge_vector_property
	template <class T>
	struct edge_vector_property : public vector_property<edge_descriptor, T>
	{
		edge_vector_property(const edge_set_base &E, const T &def = T())
			: vector_property<edge_descriptor, T>(E.num_edges(), def)
			{}

		template <class Iter>
		edge_vector_property(const edge_set_base &E, Iter begin, Iter end)
			: vector_property<edge_descriptor, T>(E.num_edges(), begin, end)
			{}
	}; // struct edge_vector_property<T>

	// shortcuts are enough for const and array properties
	template <class T>
	const_property<edge_descriptor, T>
	make_edge_property(T c) const
	{
		return const_property<edge_descriptor, T>(c);
	}

	template <class T>
	array_property<edge_descriptor, T>
	make_edge_property(T props[], size_t n) const
	{
		assert(num_edges() == n);
		return array_property<edge_descriptor, T>(props, n);
	}

	template <class T>
	array_const_property<edge_descriptor, T>
	make_edge_property(const T props[], size_t n) const
	{
		assert(num_edges() == n);
		return array_const_property<edge_descriptor, T>(props, n);
	}

	template <class T>
	array_property<edge_descriptor, T>
	make_edge_property(std::vector<T> &props) const
	{
		assert(num_edges() == props.size());
		return array_property<edge_descriptor, T>(&props[0], props.size());
	}

	template <class T>
	array_const_property<edge_descriptor, T>
	make_edge_property(const std::vector<T> &props) const
	{
		assert(num_edges() == props.size());
		return array_const_property<edge_descriptor, T>(&props[0], props.size());
	}

}; // class edge_set_base

} // namespace detail

} // namespace graph2

#endif // GRAPH2_EDGE_H
