/*
 * Copyright (c) 2003-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * graph2_property.h - Properties.
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#ifndef GRAPH2_PROPERTY_H
#define GRAPH2_PROPERTY_H

#include "../misc/shared_ptr.h"

namespace graph2
{

namespace detail
{

/*
 * Here are the concept of properties.
 * A property 'P' is a labelling on a index 'v' accessed by 'P[v]'.
 * It must be readable and could be writable.
 * The type of 'v' is P::index_type; no reference here.
 * The type of 'P[v]' is P::value_type for read and P::ref_type for write.
 *
 * All the properties here are with the reference semantics, i.e.,
 * a = b will NOT give you a copy but only a reference.
 *
 * Most of the time you don't want to use the following classes directly.
 * Look into the classes of graphs for shortcuts to types.
 */

// Property with constant value.
template <class Index, class T>
class const_property
{
	T c_;

public:
	const_property(T c) : c_(c) {}

	typedef Index index_type;
	typedef const T &value_type;

	value_type operator [] (Index) const {return c_;}

}; // class const_property<Index, T>

// Property based on a pre-existing array.
// Includes native (c style) array and vector (except std::vector<bool>).
template <class Index, class T>
class array_property
{
	T *props_;
	size_t n_;

public:
	typedef T &ref_type;
	typedef const T &value_type;
	typedef Index index_type;

	array_property(T props[], size_t n) : props_(props), n_(n) {}

	size_t size() const {return n_;}
	void fill(const T &t) {std::fill(props_, props_+n_, t);}

	template <class Iter>
	void fill(Iter begin, Iter end)
	{
		assert(size_t(std::distance(begin, end)) == n_);

		std::copy(begin, end, props_);
	}

	value_type at(index_type i) const
		{assert(i.id() < size()); return props_[i.id()];}
	ref_type at(index_type i)
		{assert(i.id() < size()); return props_[i.id()];}
	value_type operator [] (index_type i) const {return at(i);}
	ref_type operator [] (index_type i) {return at(i);}

}; // class array_property<Index, T>

template <class Index, class T>
class array_const_property
{
	const T *props_;
	size_t n_;

public:
	typedef const T &value_type;
	typedef Index index_type;

	array_const_property(const T props[], size_t n) : props_(props), n_(n) {}

	size_t size() const {return props_.size();}

	value_type at(index_type i) const
		{assert(i.id() < size()); return props_[i.id()];}
	value_type operator [] (index_type i) const {return at(i);}

}; // class array_const_property<Index, T>

// Create a property using vector.
// Save your effort of allocating memory.
// Use shared_ptr to enable reference semantics.
template <class Index, class T>
class vector_property
{
	typedef std::vector<T> vec_type;
	typedef misc::detail::shared_ptr_policy_new_delete<vec_type> policy_type;
	misc::shared_ptr<vec_type, policy_type> p_;

public:
	typedef typename vec_type::reference ref_type;
	typedef typename vec_type::const_reference value_type;
	typedef Index index_type;

	vector_property(size_t n, const T &t)
		: p_(new vec_type(n, t))
		{}

	template <class Iter>
	vector_property(size_t n, Iter begin, Iter end)
		: p_(new vec_type(begin, end))
		{assert(p_->size() == n);}

	size_t size() const {return p_->size();}
	void fill(const T &t) {std::fill(p_->begin(), p_->end(), t);}

	template <class Iter>
	void fill(Iter begin, Iter end)
	{
		assert(size_t(std::distance(begin, end)) == p_.size());

		std::copy(begin, end, p_->begin());
	}

	value_type at(index_type i) const
		{assert(i.id() < size()); return (*p_)[i.id()];}
	ref_type at(index_type i)
		{assert(i.id() < size()); return (*p_)[i.id()];}
	value_type operator [] (index_type i) const {return at(i);}
	ref_type operator [] (index_type i) {return at(i);}

}; // class vertex_property<Index, Prop>

// Properties could be built upon other properties.
// E.g., once you have 'A' and 'B', you can have 'C' such that 'C[v]=A[v]+B[v]'.
template <class Property, class Func>
class property_proxy
{
protected:
	Property &props_;
	Func func_;

public:
	typedef typename Property::index_type index_type;
	typedef typename Func::result_type value_type;

	// It's not easy to get a correct type for ref_type ...
	typedef value_type &ref_type;

	property_proxy(Property &props, Func func)
		: props_(props), func_(func) {}

	size_t size() const {return props_.size();}

	value_type at(index_type i) const {return func_(props_[i]);}
	ref_type at(index_type i) {return func_(props_[i]);}
	value_type operator [] (index_type i) const {return at(i);}
	ref_type operator [] (index_type i) {return at(i);}

}; // class property_proxy<Property, Func>

template <class Property, class Func>
class const_property_proxy
{
protected:
	const Property &props_;
	Func func_;

public:
	typedef typename Property::index_type index_type;
	typedef typename Func::result_type value_type;

	const_property_proxy(const Property &props, Func func)
		: props_(props), func_(func) {}

	size_t size() const {return props_.size();}

	value_type at(index_type i) const {return func_(props_[i]);}
	value_type operator [] (index_type i) const {return at(i);}

}; // class const_property_proxy<Property, Func>

template <class Property1, class Property2, class Func>
class const_property_proxy2
{
protected:
	const Property1 &props1_;
	const Property2 &props2_;
	Func func_;

public:
	typedef typename Property1::index_type index_type;
	typedef typename Func::result_type value_type;

	const_property_proxy2(const Property1 &props1, const Property2 &props2, Func func)
		: props1_(props1), props2_(props2), func_(func)
		{assert(props1_.size() == props2_.size());}

	size_t size() const {return props1_.size();}

	value_type at(index_type i) const {return func_(props1_[i], props2_[i]);}
	value_type operator [] (index_type i) const {return at(i);}

}; // class const_property_proxy2<Property1, Property2, Func>

} // namespace detail

template <class Property, class Func>
detail::property_proxy<Property, Func>
make_property(Property &props, Func func)
{
	return detail::property_proxy<Property, Func>(props, func);
}

template <class Property, class Func>
detail::const_property_proxy<Property, Func>
make_property(const Property &props, Func func)
{
	return detail::const_property_proxy<Property, Func>(props, func);
}

template <class Property1, class Property2, class Func>
detail::const_property_proxy2<Property1, Property2, Func>
make_property(const Property1 &props1, const Property2 &props2, Func func)
{
	return detail::const_property_proxy2<Property1, Property2, Func>(props1, props2, func);
}

} // namespace graph2

#endif // GRAPH2_PROPERTY_H
