/*
 * Copyright (c) 2003-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * graph2_v.h - Interface for vector based graph.
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#ifndef GRAPH2_V_H
#define GRAPH2_V_H

#include "../misc/common.h"
#include "graph2_property.h"
#include "graph2_vertex.h"
#include "graph2_edge.h"

namespace graph2
{

/*
 * Vector based graph
 *
 *  fixed, fast, small storage (16|E|+8|V| B)
 *
 ***************************************
 * Types
 *
 *  graph_v - the graph
 *
 *  graph_v::vertex_descriptor - type of the vertex 'v'
 *  graph_v::edge_descriptor - type of the edge 'e'
 *
 *  graph_v::edge_range - type of the range for fanins/fanouts of a vertex
 *
 *  graph_v::vertex_vector_property<T> - concrete vertex property
 *  graph_v::edge_vector_property<T> - concrete edge property
 *
 ***************************************
 * Member functions
 *
 *  graph_v(|V|, E, |E|), assign(|V|, E, |E|) - create the graph
 *
 *  num_vertices(), num_edges() - |V| and |E|
 *
 *  valid_vertices(v), valid_edge(e) - check the validity of the vertex/edge_descriptor
 *
 *  source(e), target(e) - endpoints of the edge e
 *
 *  fanins(v), fanouts(v) - range of the edges on the vertex v
 *
 *  this_edge(r) - translate the range iterator to the edge
 *
 *  make_edge_property(...), make_vertex_property(...) - connect to various existing properties
 *
 */

class graph_v : public detail::vertex_set_base, public detail::edge_set_base
{
protected:
	const size_t *edges_;

	// 2|V|+2
	// in_0 = 0, out_0 = |E|, in_1, out_1, ..., in_V = |E|, out_V = 2|E|
	std::vector<size_t> inout_base_;
	// 2|E|
	std::vector<edge_descriptor> inout_edges_;

	size_t &inbase(size_t i) {return inout_base_[i*2];}
	size_t &outbase(size_t i) {return inout_base_[i*2+1];}
	size_t inbase(size_t i) const {return inout_base_[i*2];}
	size_t outbase(size_t i) const {return inout_base_[i*2+1];}

public:
	explicit graph_v(size_t n_V = 0, const size_t edges[] = 0, size_t n_E = 0);
	void assign(size_t n_V, const size_t edges[], size_t n_E);

	vertex_descriptor source(edge_descriptor e) const
		{assert(valid_edge(e)); return vertex_descriptor(edges_[e.id()*2]);}
	vertex_descriptor target(edge_descriptor e) const
		{assert(valid_edge(e)); return vertex_descriptor(edges_[e.id()*2+1]);}

	class out_edge_iterator
	{
		friend class graph_v;
		size_t i_;

	public:
		explicit out_edge_iterator(size_t i) : i_(i) {}
		out_edge_iterator() : i_(size_t(-1)) {}

		out_edge_iterator& operator ++ () {++i_; return *this;}
		out_edge_iterator& operator -- () {--i_; return *this;}
		bool operator != (out_edge_iterator i) const {return i.i_ != i_;}
		bool operator == (out_edge_iterator i) const {return i.i_ == i_;}
	}; // class out_edge_iterator

	typedef std::pair<out_edge_iterator, out_edge_iterator> edge_range;

	edge_range fanins(vertex_descriptor v) const
		{assert(valid_vertex(v)); return std::make_pair(inbase(v.id()), inbase(v.id()+1));}
	edge_range fanouts(vertex_descriptor v) const
		{assert(valid_vertex(v)); return std::make_pair(outbase(v.id()), outbase(v.id()+1));}
	out_edge_iterator fanins_begin(vertex_descriptor v) const
		{assert(valid_vertex(v)); return out_edge_iterator(inbase(v.id()));}
	out_edge_iterator fanouts_begin(vertex_descriptor v) const
		{assert(valid_vertex(v)); return out_edge_iterator(outbase(v.id()));}
	out_edge_iterator fanins_end(vertex_descriptor v) const
		{assert(valid_vertex(v)); return out_edge_iterator(inbase(v.id()+1));}
	out_edge_iterator fanouts_end(vertex_descriptor v) const
		{assert(valid_vertex(v)); return out_edge_iterator(outbase(v.id()+1));}

	edge_descriptor this_edge(out_edge_iterator oi) const
		{assert(oi.i_ < 2*num_edges()); return inout_edges_[oi.i_];}

	size_t num_fanins(vertex_descriptor v) const
		{assert(valid_vertex(v)); return inbase(v.id()+1)-inbase(v.id());}
	size_t num_fanouts(vertex_descriptor v) const
		{assert(valid_vertex(v)); return outbase(v.id()+1)-outbase(v.id());}

	// direct access of edge list
	typedef std::vector<edge_descriptor>::iterator direct_iterator;
	typedef std::pair<direct_iterator, direct_iterator> direct_range;
	out_edge_iterator to_edge_iterator(direct_iterator di)
		{return out_edge_iterator(di-inout_edges_.begin());}
	direct_iterator to_direct_iterator(out_edge_iterator ei)
		{return inout_edges_.begin()+ei.i_;}
	edge_range to_edge_range(direct_range dr)
		{return std::make_pair(to_edge_iterator(dr.first), to_edge_iterator(dr.second));}
	direct_range to_direct_range(edge_range er)
		{return std::make_pair(to_direct_iterator(er.first), to_direct_iterator(er.second));}

	// templated typedef for properties
	template<class T>
	struct vertex_properties
	{
		typedef detail::const_property<vertex_descriptor, T> const_prop;
		typedef detail::array_property<vertex_descriptor, T> array_prop;
		typedef detail::array_const_property<vertex_descriptor, T> array_const_prop;
		typedef vertex_vector_property<T> vector_prop;
	}; // struct vertex_properties<T>

	template<class T>
	struct edge_properties
	{
		typedef detail::const_property<edge_descriptor, T> const_prop;
		typedef detail::array_property<edge_descriptor, T> array_prop;
		typedef detail::array_const_property<edge_descriptor, T> array_const_prop;
		typedef edge_vector_property<T> vector_prop;
	}; // struct edge_properties<T>

}; // class graph_v

} // namespace graph2

#endif // GRAPH2_V_H
