/*
 * Copyright (c) 2004-2007 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * graph_utilities.h - Useful basic data structure for graph traversal.
 *
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#ifndef GRAPH_UTILITIES_H
#define GRAPH_UTILITIES_H

namespace graph
{

// fast coloring without the need to clear everytime before use
class coloring
{
	std::vector<size_t> colors_;
	size_t lb_, ub_; // [0, ub) are valid, [0, lb) are not colored

public:
	coloring() : lb_(0), ub_(1) {}
	explicit coloring(size_t n_v) : colors_(n_v, 0), lb_(0), ub_(1) {}
	void assign(size_t n_v) {colors_.assign(n_v, 0); lb_ = 0; ub_ = 1;}

	size_t size() const {return colors_.size();}

	void reset_all(size_t ub = 1)
	{
		if (ub_+ub < ub_) // overflow
		{
			lb_ = 1;
			std::fill(colors_.begin(), colors_.end(), 0);
		}
		else
		{
			lb_ = ub_;
		}
		ub_ = lb_+ub;

		assert(ub_ > lb_);
	}

	void set_color(size_t v, size_t c = 0)
	{
		assert(v < size());
		assert(lb_ != 0);
		assert(colors_[v] < ub_);
		assert(c < ub_-lb_);

		colors_[v] = lb_+c;
	}

	void reset_color(size_t v)
	{
		assert(v < size());
		assert(lb_ != 0);
		assert(colors_[v] < ub_);

		colors_[v] = lb_-1;
	}

	bool colored(size_t v) const
	{
		assert(v < size());
		assert(lb_ != 0);
		assert(colors_[v] < ub_);

		return colors_[v] >= lb_;
	}

	// only call this when you know it is colored
	size_t get_color(size_t v) const
	{
		assert(colored(v));

		return colors_[v]-lb_;
	}

}; // class coloring

// fast queue, it's NOT a circular queue
// but can be used with coloring for BFS/DFS
template <class T>
class fast_queue
{
	std::vector<T> queue_;
	size_t begin_, end_;

public:
	fast_queue() : begin_(0), end_(0) {}
	explicit fast_queue(size_t n_v) : queue_(n_v), begin_(0), end_(0) {}
	void assign(size_t n_v) {queue_.resize(n_v); begin_ = end_ = 0;}

	void clear() {begin_ = end_ = 0;}

	size_t size() const {return end_-begin_;}
	bool empty() const {return begin_ == end_;}
	bool front_full() const {return begin_ == 0;}
	bool back_full() const {return end_ == queue_.size();}

	void pop_front() {assert(!empty()); begin_++;}
	void pop_back() {assert(!empty()); --end_;}
	void push_front(const T &v) {assert(!front_full()); queue_[--begin_] = v;}
	void push_back(const T &v) {assert(!back_full()); queue_[end_++] = v;}

	const T &front() const {assert(!empty()); return queue_[begin_];}
	const T &back() const {assert(!empty()); return queue_[end_-1];}
	T &front() {assert(!empty()); return queue_[begin_];}
	T &back() {assert(!empty()); return queue_[end_-1];}

	// for accessing queue directly
	typedef typename std::vector<T>::const_iterator iter;
	typedef std::pair<iter, iter> range;
	range history() const
		{return range(queue_.begin(), queue_.begin()+begin_);}
	range queue() const
		{return range(queue_.begin()+begin_, queue_.begin()+end_);}
	iter begin() const {return queue_.begin()+begin_;}
	iter end() const {return queue_.begin()+end_;}

}; // class fast_queue<T>

template <class T>
class circular_queue
{
	std::vector<T> queue_;
	size_t begin_, end_;

	size_t prev(size_t i)
	{
		if (i != 0)
			return i-1;
		else
			return queue_.size()-1;
	}

	size_t next(size_t i)
	{
		++i;
		if (i == queue_.size())
			i = 0;
		return i;
	}

public:
	circular_queue() : queue_(1), begin_(0), end_(0) {}
	explicit circular_queue(size_t n) : queue_(n+1), begin_(0), end_(0) {}
	void assign(size_t n) {queue_.resize(n+1); begin_ = end_ = 0;}

	size_t size() const
	{
		if (begin_ <= end_)
			return end_-begin_;
		else
			return queue_.size()+end_-begin_;
	}

	bool empty() const {return begin_ == end_;}

	bool full() const {return size() == queue_.size()-1;}

	const T &front() const {assert(!empty()); return queue_[begin_];}
	const T &back() const {assert(!empty()); return queue_[prev(end_)];}
	T &front() {assert(!empty()); return queue_[begin_];}
	T &back() {assert(!empty()); return queue_[prev(end_)];}

	void push_front(const T &v)
	{
		assert(!full());

		begin_ = prev(begin_);
		queue_[begin_] = v;
	}

	void push_back(const T &v)
	{
		assert(!full());

		queue_[end_] = v;
		end_ = next(end_);
	}

	void pop_front() {assert(!empty()); begin_ = next(begin_);}
	void pop_back() {assert(!empty()); end_ = prev(end_);}

}; // class circular_queue<T>

// fast list represents a subset of {0,1,...,n-1}
// insert/erase/check each takes O(1) time
class fast_list
{
	struct list_entry
	{
		size_t prev_id, next_id;
	}; // struct list_entry

	std::vector<list_entry> elements_;
	size_t size_;

public:
	fast_list() : size_(0) {}
	explicit fast_list(size_t n_v) : size_(0) {assign(n_v);}
	void assign(size_t n_v) {elements_.resize(n_v+1); erase_all();}

	size_t size() const {return size_;}
	bool empty() const {return size_ == 0;}

	void insert_all()
	{
		size_ = elements_.size()-1;
		for (size_t i = 0; i <= size_; ++i)
		{
			elements_[i].prev_id = i-1;
			elements_[i].next_id = i+1;
		}
		elements_[0].prev_id = size_;
		elements_[size_].next_id = 0;
	}

	void erase_all()
	{
		size_ = 0;
		for (size_t i = 0; i < elements_.size(); ++i)
			elements_[i].prev_id = elements_[i].next_id = i;
	}

	bool in_list(size_t i) const
	{
		assert(i+1 < elements_.size());
		return elements_[i].next_id != i;
	}

	// return the element after i
	size_t erase(size_t i)
	{
		assert(in_list(i));
		size_t p = elements_[i].prev_id;
		size_t n = elements_[i].next_id;
		elements_[p].next_id = n;
		elements_[n].prev_id = p;
		elements_[i].prev_id = elements_[i].next_id = i;
		--size_;
		return n;
	}

	void insert(size_t i, size_t after)
	{
		assert(!in_list(i));
		assert((after == elements_.size()-1) || in_list(after));
		size_t p = elements_[after].prev_id;
		elements_[p].next_id = i;
		elements_[after].prev_id = i;
		elements_[i].next_id = after;
		elements_[i].prev_id = p;
		++size_;
	}

	size_t front() const {assert(!empty()); return elements_.back().next_id;}
	size_t back() const {assert(!empty()); return elements_.back().prev_id;}

	size_t begin() const {return elements_.back().next_id;}
	size_t end() const {return elements_.size()-1;}

	size_t next(size_t i) const {assert(in_list(i)); return elements_[i].next_id;}
	size_t prev(size_t i) const {assert(in_list(i)); return elements_[i].prev_id;}

	size_t pop_front() {size_t i = front(); erase(i); return i;}
	size_t pop_back() {size_t i = back(); erase(i); return i;}
	void push_front(size_t i) {insert(i, begin());}
	void push_back(size_t i) {insert(i, end());}

	void move_front(size_t i){erase(i); push_front(i);}
	void move_back(size_t i){erase(i); push_back(i);}

}; // class fast_list

// helps sorting edges
template <class Iter, class Head, class Tail> std::pair<Iter, Iter>
triple_partition(Iter first, Iter last, Head head, Tail tail)
{
	/*
	 * [begin, first) head
	 * [first, it) mid
	 * [it, last) unknown
	 * [last, end) tail
	 */
	for (Iter it = first; it != last;)
		if (head(*it))
		{
			if (it != first)
				std::swap(*it, *first);
			++first;
			++it;
		}
		else if (tail(*it))
		{
			for (--last; it != last; --last)
				if (!tail(*last))
					break;
			if (it != last)
				std::swap(*it, *last);
		}
		else
		{
			++it;
		}

	return std::make_pair(first, last);
}

} // namespace graph

#endif // GRAPH_UTILITIES_H
