/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * hyper_graph.h - Hyper-graph.
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#ifndef HYPER_GRAPH_H
#define HYPER_GRAPH_H

namespace graph
{

/*
 * vector based hyper-graph
 *
 * Hyper-graph is a set of edges where every edge is a set of elements.
 * The sets are implemented via vectors (arrays) such that the storage
 * requirement is O(|V|+|E|) while random accessing and appending take
 * O(1) time. General insertion is allowed but takes O(|V|+|E|) time.
 *
 * U: edge property
 * V: element property
 *
 * You can acess the properties by the 'at' function or by [] like an
 * array or by () like a matrix. An alternative way is to use the
 * iterators.
 *
 * Generally speaking, enumerating a linear data structure (array/vector,
 * list, ordered binary tree, etc) is done via iterators (like pointers
 * to the array). The range is half close and half open, i.e., [begin, end)
 * where 'begin' and 'end' are both iterators.
 *
 * Iterators are designed to be working 'as' pointers - that means '*'
 * and '->' are both valid and most of the time, '++' and '--' can be
 * used.
 */
template <class U, class T>
class hyper_graph_v
{
	std::vector<T> elements_;
	std::vector<size_t> offsets_;
	std::vector<U> edges_;

public:
	hyper_graph_v() {offsets_.push_back(0);}

	void clear()
	{
		offsets_.resize(1);
		offsets_[0] = 0;

		elements_.resize(0);
		
		edges_.resize(0);
	}

	void reserve(size_t n_edges, size_t n_elements)
	{
		elements_.reserve(n_elements);
		offsets_.reserve(n_edges+1);
		edges_.reserve(n_edges);
	}

	size_t num_edges() const {return edges_.size();}
	size_t num_elements() const {return elements_.size();}

	// # elements per edge
	size_t num_elements(size_t i_edge) const
	{
		assert(i_edge < num_edges());
		return offsets_[i_edge+1]-offsets_[i_edge];
	}

	// accesser
	U &at(size_t i_edge)
	{
		assert(i_edge < num_edges());

		return edges_[i_edge];
	}

	const U &at(size_t i_edge) const
	{
		assert(i_edge < num_edges());

		return edges_[i_edge];
	}

	T &at(size_t i_edge, size_t i_element)
	{
		assert(i_edge < num_edges());
		assert(i_element < num_elements());

		return elements_[offsets_[i_edge]+i_element];
	}

	const T &at(size_t i_edge, size_t i_element) const
	{
		assert(i_edge < num_edges());
		assert(i_element < num_elements());

		return elements_[offsets_[i_edge]+i_element];
	}

	U &operator [] (size_t i_edge) {return at(i_edge);}
	const U &operator [] (size_t i_edge) const {return at(i_edge);}

	T &operator () (size_t i_edge, size_t i_element) {return at(i_edge, i_element);}
	const T &operator () (size_t i_edge, size_t i_element) const {return at(i_edge, i_element);}

	// iterators for edge properties
	typedef typename std::vector<U>::iterator iterator;
	typedef typename std::vector<U>::const_iterator const_iterator;

	iterator begin() {return edges_.begin();}
	iterator end() {return edges_.end();}

	const_iterator begin() const {return edges_.begin();}
	const_iterator end() const {return edges_.end();}

	// iterators for element properties
	typedef typename std::vector<T>::iterator element_iterator;
	typedef typename std::vector<T>::const_iterator const_element_iterator;

	// elements in individual edges
	element_iterator el_begin(size_t i_edge)
	{
		assert(i_edge < num_edges());
		return elements_.begin()+offsets_[i_edge];
	}
	element_iterator el_end(size_t i_edge)
	{
		assert(i_edge < num_edges());
		return elements_.begin()+offsets_[i_edge+1];
	}

	const_element_iterator el_begin(size_t i_edge) const
	{
		assert(i_edge < num_edges());
		return elements_.begin()+offsets_[i_edge];
	}
	const_element_iterator el_end(size_t i_edge) const
	{
		assert(i_edge < num_edges());
		return elements_.begin()+offsets_[i_edge+1];
	}

	// all the elements
	element_iterator el_begin()
		{return elements_.begin();}
	const_element_iterator el_begin() const
		{return elements_.begin();}
	element_iterator el_end()
		{return elements_.end();}
	const_element_iterator el_end() const
		{return elements_.end();}

	// range for element properties
	typedef std::pair<element_iterator, element_iterator> range;
	typedef std::pair<const_element_iterator, const_element_iterator> const_range;
	typedef range range_t;
	typedef const_range const_range_t;

	// elements in individual edges
	range_t elements(size_t i_edge)
		{return range_t(el_begin(i_edge), el_end(i_edge));}
	const_range_t elements(size_t i_edge) const
		{return const_range_t(el_begin(i_edge), el_end(i_edge));}
	// all the elements
	range_t elements()
		{return range_t(elements_.begin(), elements_.end());}
	const_range_t elements() const
		{return const_range_t(elements_.begin(), elements_.end());}

	// operations
	void append_edge(const U &u = U())
	{
		offsets_.push_back(elements_.size());
		edges_.push_back(u);
	}

	void append_element(const T &t = T())
	{
		elements_.push_back(t);
		++offsets_.back();
	}

	void swap(hyper_graph_v<U, T> &r)
	{
		elements_.swap(r.elements_);
		offsets_.swap(r.offsets_);
		edges_.swap(r.edges_);
	}

	// slow operations ... but we need them

	void insert_edge(size_t i, const U &u = U())
	{
		edges_.insert(edges_.begin()+i, u);
		offsets_.insert(offsets_.begin()+i+1, offsets_[i]);
	}

	void append_element(size_t i_edge, const T &t = T())
	{
		elements_.insert(el_end(i_edge), t);
		for (size_t i = i_edge+1; i < offsets_.size(); ++i)
			++offsets_[i];
	}

	// just in case you understand the data structure and want
	// something more ... (like transpose)
	std::vector<T> &get_internal_elements() {return elements_;}
	std::vector<size_t> &get_internal_offsets() {return offsets_;}
	std::vector<U> &get_internal_edges() {return edges_;}

}; // class hyper_graph_v<T>

} // namespace graph

#endif // HYPER_GRAPH_H
